/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <aul.h>
#include <aul_rpc_port.h>
#include <dlog.h>

#include "fdbroker-internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "RPC_PORT"

#define EILLEGALACCESS 127

#define DBUS_SERVICE_DBUS "org.freedesktop.DBus"
#define DBUS_PATH_DBUS "/org/freedesktop/DBus"
#define DBUS_INTERFACE_DBUS "org.freedesktop.DBus"
#define RPC_PORT_OBJECT_PATH "/org/tizen/rpcport"
#define RPC_PORT_INTERFACE_PREFIX "org.tizen.rpcport._"

namespace rpc_port {
namespace internal {

FdBroker::DBusConnectionManager& FdBroker::DBusConnectionManager::GetInst() {
  static DBusConnectionManager dm;

  if (dm.disposed_) {
    dm.Init();
  }

  return dm;
}

void FdBroker::DBusConnectionManager::Dispose() {
  if (!disposed_)
    Fini();
}

GDBusConnection* FdBroker::DBusConnectionManager::GetConnection() {
  return gdbus_conn_;
}

FdBroker::DBusConnectionManager::~DBusConnectionManager() {
  if (!disposed_)
    Fini();
}

void FdBroker::DBusConnectionManager::Init() {
  GError* error = nullptr;

  gdbus_conn_ = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, &error);
  if (gdbus_conn_ == nullptr) {
    if (error != nullptr) {
      LOGE("Failed to get dbus [%s]", error->message);
      g_error_free(error);
    }

    return;
  }

  disposed_ = false;
}

void FdBroker::DBusConnectionManager::Fini() {
  if (gdbus_conn_ != nullptr) {
    g_object_unref(gdbus_conn_);
    gdbus_conn_ = nullptr;
  }

  disposed_ = true;
}

FdBroker::DBusMock& FdBroker::DBusMock::GetInst() {
  static DBusMock mock;

  return mock;
}

int FdBroker::DBusMock::Send(const std::string& sender,
                             const std::string& port, int fds[2]) {
  if (port == "wrong_port")
    return -EILLEGALACCESS;
  if (ports_.find(port) == ports_.end())
    return -1;

  ports_[port]->OnFdReceived(sender, fds);
  return 0;
}

int FdBroker::DBusMock::AddListener(const std::string& port,
                                    FdBroker::IEventListener* listener) {
  if (ports_.find(port) != ports_.end())
    return -1;

  ports_[port] = listener;
  return 0;
}

int FdBroker::DBusMock::Watch(FdBroker::IEventWatcher* watcher,
                              const std::string& target_appid,
                              const std::string& port_name) {
  watcher->OnPortAppeared(target_appid, port_name);
  return 0;
}

void FdBroker::DBusMock::Dispose() {
  ports_.clear();
}

FdBroker::SocketPair::SocketPair(bool mock)
    : mock_(mock) {
  socks_[SENDER] = 0;
  socks_[RECEIVER] = 0;
}

FdBroker::SocketPair::~SocketPair() {
  if (socks_[SENDER] > 0)
    close(socks_[SENDER]);
  if (socks_[RECEIVER] > 0)
    close(socks_[RECEIVER]);
}

int FdBroker::SocketPair::Request(const std::string& target_appid,
                                  const std::string& port_name) {
  if (mock_) {
    return socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, socks_);
  }

  if (aul_rpc_port_create_socket_pair(target_appid.c_str(),
      port_name.c_str(), &socks_) != AUL_R_OK) {
    LOGE("error create socket pair");
    return -1;
  }

  return 0;
}

int FdBroker::SocketPair::Get(Type t) const {
  return socks_[t];
}

int FdBroker::SocketPair::Detach(Type t) {
  int fd = socks_[t];

  socks_[t] = 0;
  return fd;
}

FdBroker::FdList::FdList() {
  fd_list_ = g_unix_fd_list_new();
}

FdBroker::FdList::~FdList() {
  g_object_unref(fd_list_);
}

int FdBroker::FdList::Add(int fd) {
  GError *err = nullptr;

  g_unix_fd_list_append(fd_list_, fd, &err);
  close(fd);

  if (err != NULL) {
    LOGE("g_unix_fd_list_append [%s]", err->message);
    g_error_free(err);
    return -1;
  }

  return 0;
}

GUnixFDList* FdBroker::FdList::GetRaw() {
  return fd_list_;
}

FdBroker::~FdBroker() {
  if (cancellable_) {
    LOGW("Cancel the send request");
    g_cancellable_cancel(cancellable_);
    g_object_unref(cancellable_);
  }

  if (watcher_id_ > 0)
    g_bus_unwatch_name(watcher_id_);

  if (registration_id_ > 0) {
    g_dbus_connection_unregister_object(
        DBusConnectionManager::GetInst().GetConnection(),
        registration_id_);
  }

  if (mock_) {
    DBusMock::GetInst().Dispose();
  }
}

std::string FdBroker::GetInterfaceName(const std::string& target_appid,
                             const std::string& port_name) {
  std::string interface_name = target_appid + "_" + port_name;
  char c_buf[interface_name.length() * 2 + 1];
  char* temp = &c_buf[0];
  memset(c_buf, 0, sizeof(c_buf));
  for (unsigned int index = 0; index < interface_name.length(); index++) {
    snprintf(temp, 3, "%02x", interface_name[index]);
    temp += 2;
  }

  return RPC_PORT_INTERFACE_PREFIX + std::string(c_buf);
}

int FdBroker::Send(const std::string& target_appid,
                   const std::string& port_name,
                   int (*fds)[2]) {
  std::string interface_name = GetInterfaceName(target_appid, port_name);
  GDBusMessage *msg;
  SocketPair main_sock_pair(mock_);
  SocketPair delegate_sock_pair(mock_);
  FdList fd_list;
  char sender_appid[255];

  if (!mock_ && aul_app_get_appid_bypid(getpid(),
       sender_appid, sizeof(sender_appid)) < 0) {
    LOGE("Can't get appid");
    return -1;
  }

  if (main_sock_pair.Request(target_appid, port_name) != 0)
    return -1;
  if (delegate_sock_pair.Request(target_appid, port_name) != 0)
    return -1;

  if (mock_) {
    int send_fds[2];
    send_fds[0] = main_sock_pair.Detach(SocketPair::RECEIVER);
    send_fds[1] = delegate_sock_pair.Detach(SocketPair::RECEIVER);
    int ret = DBusMock::GetInst().Send("TestApp", port_name, send_fds);
    if (ret < 0)
      return ret;

    (*fds)[0] = main_sock_pair.Detach(SocketPair::SENDER);
    (*fds)[1] = delegate_sock_pair.Detach(SocketPair::SENDER);
    return (*fds)[0];
  }

  if (fd_list.Add(main_sock_pair.Detach(SocketPair::RECEIVER)) != 0)
    return -1;
  if (fd_list.Add(delegate_sock_pair.Detach(SocketPair::RECEIVER)) != 0)
    return -1;

  msg = g_dbus_message_new_method_call(interface_name.c_str(),
                                       RPC_PORT_OBJECT_PATH,
                                       interface_name.c_str(), "send_message");
  if (!msg) {
    LOGE("Can't allocate new method call");
    return -1;
  }

  if (cancellable_) {
    g_cancellable_cancel(cancellable_);
    g_object_unref(cancellable_);
  }

  cancellable_ = g_cancellable_new();
  if (!cancellable_) {
    LOGE("Failed to create GCancellable");
    g_object_unref(msg);
    return -1;
  }

  g_dbus_message_set_unix_fd_list(msg, fd_list.GetRaw());
  g_dbus_connection_send_message_with_reply(
      DBusConnectionManager::GetInst().GetConnection(),
      msg, G_DBUS_SEND_MESSAGE_FLAGS_NONE,
      5000, NULL, cancellable_, OnResultReceived, this);
  g_object_unref(msg);

  (*fds)[0] = main_sock_pair.Detach(SocketPair::SENDER);
  (*fds)[1] = delegate_sock_pair.Detach(SocketPair::SENDER);

  return (*fds)[0];
}

void FdBroker::ReceiveMessage(const char* sender_appid,
                              GDBusMethodInvocation* invocation) {
  GDBusMessage* msg;
  GUnixFDList* fd_list;
  int fd_len = 0;
  int* returned_fds = nullptr;

  msg = g_dbus_method_invocation_get_message(invocation);
  fd_list = g_dbus_message_get_unix_fd_list(msg);

  if (fd_list == nullptr)
    return;

  returned_fds = g_unix_fd_list_steal_fds(fd_list, &fd_len);
  if (returned_fds == nullptr || fd_len != 2) {
    LOGE("fail to get fds. fd_len(%d)", fd_len);
    return;
  }

  listener_->OnFdReceived(sender_appid, returned_fds);
  free(returned_fds);
}

void FdBroker::OnReceiveDbusMethod(GDBusConnection *conn,
    const gchar *sender, const gchar *object_path,
    const gchar *iface_name, const gchar *method_name,
    GVariant *parameters, GDBusMethodInvocation *invocation,
    gpointer user_data) {
  FdBroker* broker = static_cast<FdBroker*>(user_data);
  int ret = -1;
  char sender_appid[255];
  int sender_pid;

  sender_pid = broker->GetSenderPid(conn, sender);
  if (aul_app_get_appid_bypid(sender_pid, sender_appid,
      sizeof(sender_appid)) < 0) {
    g_dbus_method_invocation_return_value(invocation,
        g_variant_new("(i)", ret));
    return;
  }

  AccessController& ac = broker->GetAccessController();
  ret = ac.Check(conn, sender, sender_appid);
  if (ret == 0)
    broker->ReceiveMessage(sender_appid, invocation);

  g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

int FdBroker::GetOwnerId(const std::string& interface_name) {
  int owner_id = 0;
  GError *error = NULL;

  GVariant* result = g_dbus_connection_call_sync(
      DBusConnectionManager::GetInst().GetConnection(),
      DBUS_SERVICE_DBUS,
      DBUS_PATH_DBUS,
      DBUS_INTERFACE_DBUS,
      "RequestName",
      g_variant_new("(su)", interface_name.c_str(),
          G_BUS_NAME_OWNER_FLAGS_NONE),
      G_VARIANT_TYPE("(u)"),
      G_DBUS_CALL_FLAGS_NONE,
      -1,
      nullptr,
      &error);

  if (error) {
    LOGE("RequestName fail : %s", error->message);
    g_error_free(error);
    return -1;
  }

  if (result == nullptr) {
    LOGE("fail to get name NULL");
    return -1;
  }

  g_variant_get(result, "(u)", &owner_id);
  if (owner_id == 0) {
    LOGE("Acquiring the own name is failed");
    g_variant_unref(result);
    return -1;
  }

  LOGD("Acquiring the own name : %d", owner_id);
  g_variant_unref(result);

  return owner_id;
}

int FdBroker::GetSenderPid(GDBusConnection *connection, const gchar *sender) {
  GDBusMessage *msg = NULL;
  GDBusMessage *reply = NULL;
  GError *err = NULL;
  GVariant *body;
  int pid = 0;

  msg = g_dbus_message_new_method_call("org.freedesktop.DBus",
      "/org/freedesktop/DBus", "org.freedesktop.DBus",
      "GetConnectionUnixProcessID");
  if (!msg) {
    LOGE("Can't allocate new method call");
    goto out;
  }

  g_dbus_message_set_body(msg, g_variant_new("(s)", sender));
  reply = g_dbus_connection_send_message_with_reply_sync(connection, msg,
      G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &err);

  if (!reply) {
    if (err != NULL) {
      LOGE("Failed to get pid [%s]", err->message);
      g_error_free(err);
    }
    goto out;
  }

  body = g_dbus_message_get_body(reply);
  g_variant_get(body, "(u)", &pid);

out:
  if (msg)
    g_object_unref(msg);
  if (reply)
    g_object_unref(reply);

  return pid;
}

int FdBroker::RegisterDbusInterface(const std::string& port_name) {
  static const GDBusInterfaceVTable interface_vtable = {
    OnReceiveDbusMethod,
    nullptr,
    nullptr
  };
  static const char introspection_prefix[] =
    "<node>"
    "  <interface name='";
  static const char introspection_postfix[] =
    "'>"
    "  <method name='send_message'>"
    "    <arg type='i' name='response' direction='out'/>"
    "  </method>"
    "  </interface>"
    "</node>";
  char appid[255];

  if (aul_app_get_appid_bypid(getpid(), appid, sizeof(appid)) < 0)
    return -1;

  std::string interface_name = GetInterfaceName(appid, port_name);
  std::string introspection_xml = introspection_prefix +
                                  interface_name +
                                  introspection_postfix;

  GDBusNodeInfo* introspection_data = g_dbus_node_info_new_for_xml(
      introspection_xml.c_str(), nullptr);
  if (!introspection_data) {
    LOGE("g_dbus_node_info_new_for_xml() is failed.");
    return -1;
  }

  registration_id_ = g_dbus_connection_register_object(
      DBusConnectionManager::GetInst().GetConnection(),
      RPC_PORT_OBJECT_PATH, introspection_data->interfaces[0],
      &interface_vtable, this, nullptr, nullptr);

  g_dbus_node_info_unref(introspection_data);
  if (registration_id_ == 0) {
    LOGE("Failed to g_dbus_connection_register_object");
    return -1;
  }

  if (GetOwnerId(interface_name) < 0) {
    LOGE("Failed to get owner id");
    return -1;
  }

  LOGI("%s is registered", port_name.c_str());

  return 0;
}

int FdBroker::Listen(IEventListener* ev, const std::string& port_name) {
  if (listener_ != nullptr) {
    LOGE("listener_ is not NULL");
    return RPC_PORT_ERROR_INVALID_PARAMETER;
  }

  if (ev == nullptr) {
    LOGE("ev is NULL");
    return RPC_PORT_ERROR_INVALID_PARAMETER;
  }

  if (mock_) {
    int ret = DBusMock::GetInst().AddListener(port_name, ev);

    if (ret < 0)
      return ret;

    return RPC_PORT_ERROR_NONE;
  }

  int ret = RegisterDbusInterface(port_name);

  if (ret != 0) {
    LOGE("Failed to register dbus interface");
    return RPC_PORT_ERROR_IO_ERROR;
  }

  listener_ = ev;
  return RPC_PORT_ERROR_NONE;
}

AccessController& FdBroker::GetAccessController() {
  return ac_;
}

void FdBroker::OnNameAppeared(GDBusConnection *connection,
                              const gchar *name,
                              const gchar *name_owner,
                              gpointer user_data) {
  FdBroker* broker = static_cast<FdBroker*>(user_data);
  int pid = broker->GetSenderPid(connection, name_owner);
  char owner_appid[255] = { 0, };

  if (aul_app_get_appid_bypid(pid, owner_appid, sizeof(owner_appid)) < 0) {
    LOGE("aul_app_get_appid_bypid failed %d", pid);
    broker->watcher_->OnPortRejected(owner_appid);
    return;
  }

  if (broker->watch_appid_ != owner_appid) {
    LOGE("invalid appid %s", owner_appid);
    broker->watcher_->OnPortRejected(owner_appid);
    return;
  }

  broker->watcher_->OnPortAppeared(broker->watch_appid_,
      broker->watch_port_name_);
}

void FdBroker::OnNameVanished(GDBusConnection *connection,
                              const gchar *name,
                              gpointer user_data) {
  FdBroker* broker = static_cast<FdBroker*>(user_data);
  broker->watcher_->OnPortVanished(broker->watch_appid_,
      broker->watch_port_name_);
}

int FdBroker::Watch(IEventWatcher* ev, const std::string& target_appid,
                    const std::string& port_name) {
  int r;

  if (ev == nullptr)
    return -1;

  if (!mock_) {
    r = aul_rpc_port_prepare_stub(target_appid.c_str(), port_name.c_str());
    if (r != AUL_R_OK) {
      LOGE("Failed to prepare stub %s:%s [ret : %d]",
          target_appid.c_str(), port_name.c_str(), r);
      return r;
    }
  }

  watcher_ = ev;
  watch_appid_ = target_appid;
  watch_port_name_ = port_name;

  if (mock_) {
    r = DBusMock::GetInst().Watch(ev, target_appid, port_name);
    if (r < 0)
      return -1;

    return 0;
  }

  std::string interface_name = GetInterfaceName(target_appid, port_name);

  if (watcher_id_ > 0) {
    g_bus_unwatch_name(watcher_id_);
    LOGD("Retry to watch name. stub %s:%s",
        target_appid.c_str(), port_name.c_str());
  }

  watcher_id_ = g_bus_watch_name_on_connection(
      DBusConnectionManager::GetInst().GetConnection(),
      interface_name.c_str(),
      G_BUS_NAME_WATCHER_FLAGS_NONE,
      OnNameAppeared,
      OnNameVanished,
      this,
      NULL);
  if (watcher_id_ == 0) {
    LOGE("Failed to watch connection(%s)", interface_name.c_str());
    watcher_ = nullptr;
    return -1;
  }

  return 0;
}

void FdBroker::OnResultReceived(GObject* source_object,
                                GAsyncResult* res,
                                gpointer user_data) {
  GDBusConnection* conn = reinterpret_cast<GDBusConnection*>(source_object);
  GError* err = nullptr;
  GDBusMessage* reply = g_dbus_connection_send_message_with_reply_finish(conn,
      res, &err);
  if (reply == nullptr) {
    LOGE("No reply. err(%s)", err ? err->message : "Unknown");
    if (err && err->code == G_IO_ERROR_CANCELLED) {
      g_error_free(err);
      return;
    }
  }

  FdBroker* broker = static_cast<FdBroker*>(user_data);
  IEventWatcher* watcher = broker->watcher_;

  if (err) {
    watcher->OnPortDisconnected(broker->watch_appid_, broker->watch_port_name_,
        true);
    g_error_free(err);
    return;
  }

  GVariant* reply_body = g_dbus_message_get_body(reply);
  if (reply_body == nullptr) {
    LOGE("g_dbus_message_get_body() is failed");
    watcher->OnPortDisconnected(broker->watch_appid_, broker->watch_port_name_);
    g_object_unref(reply);
    return;
  }

  int ret;
  g_variant_get(reply_body, "(i)", &ret);
  g_object_unref(reply);
  if (ret != 0) {
    LOGE("Access Denied[sender_appid : %s]", broker->watch_appid_.c_str());
    watcher->OnPortRejected(broker->watch_appid_);
    return;
  }

  watcher->OnPortConnected(broker->watch_appid_, broker->watch_port_name_);
  LOGD("[Reply : %d]", ret);
}

}  // namespace internal
}  // namespace rpc_port
