/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef FDBROKER_INTERNAL_H_
#define FDBROKER_INTERNAL_H_

#include <glib.h>
#include <gio/gio.h>
#include <gio/gunixfdlist.h>
#include <glib-unix.h>

#include <string>
#include <map>
#include <memory>

#include "ac-internal.h"
#include "rpc-port.h"

namespace rpc_port {
namespace internal {

class FdBroker {
 public:
  class IEventListener {
   public:
    virtual void OnFdReceived(const std::string& sender, int fds[2]) = 0;
  };

  class IEventWatcher {
   public:
    virtual void OnPortAppeared(const std::string& appid,
                                const std::string& port_name) = 0;
    virtual void OnPortVanished(const std::string& appid,
                                const std::string& port_name) = 0;
    virtual void OnPortRejected(const std::string& appid) = 0;
    virtual void OnPortConnected(const std::string& appid,
                                 const std::string& port_name) = 0;
    virtual void OnPortDisconnected(const std::string& appid,
                                    const std::string& port_name,
                                    bool cancel = false) = 0;
  };

  explicit FdBroker(bool mock = false) : mock_(mock) {}
  ~FdBroker();

  static void Dispose() {
    DBusConnectionManager::GetInst().Dispose();
  }

  int Send(const std::string& target_appid, const std::string& port_name,
           int (*fds)[2]);
  int Listen(IEventListener* ev, const std::string& port_name);
  AccessController& GetAccessController();
  int Watch(IEventWatcher* ev, const std::string& target_appid,
            const std::string& port_name);

 private:
  class DBusConnectionManager {
   public:
    DBusConnectionManager(const DBusConnectionManager&) = delete;
    DBusConnectionManager& operator = (const DBusConnectionManager&) = delete;

    static DBusConnectionManager& GetInst();
    void Dispose();
    GDBusConnection* GetConnection();

   private:
    DBusConnectionManager() = default;
    ~DBusConnectionManager();

    void Init();
    void Fini();

   private:
    bool disposed_ = true;
    GDBusConnection* gdbus_conn_ = nullptr;
  };

  class DBusMock {
   public:
    DBusMock(const DBusMock&) = delete;
    DBusMock& operator = (const DBusMock&) = delete;

    static DBusMock& GetInst();
    int Send(const std::string& sender, const std::string& port, int fds[2]);
    int AddListener(const std::string& port,
                    FdBroker::IEventListener* listener);
    int Watch(FdBroker::IEventWatcher* watcher, const std::string& target_appid,
              const std::string& port_name);
    void Dispose();

   private:
    DBusMock() = default;
    ~DBusMock() = default;

   private:
    std::map<std::string, FdBroker::IEventListener*> ports_;
  };

  class SocketPair {
   public:
    enum Type {
      SENDER = 0,
      RECEIVER = 1
    };

    explicit SocketPair(bool mock = false);
    ~SocketPair();

    int Request(const std::string& target_appid, const std::string& port_name);
    void RequestMock();
    int Get(Type t) const;
    int Detach(Type t);

   private:
    int socks_[2];
    bool mock_;
  };

  class FdList {
   public:
    FdList();
    ~FdList();

    int Add(int fd);
    GUnixFDList* GetRaw();

   private:
    GUnixFDList* fd_list_;
  };

 private:
  static void OnReceiveDbusMethod(GDBusConnection *conn, const gchar *sender,
                                  const gchar *object_path,
                                  const gchar *iface_name,
                                  const gchar *method_name,
                                  GVariant *parameters,
                                  GDBusMethodInvocation *invocation,
                                  gpointer user_data);
  int GetOwnerId(const std::string& interface_name);
  int GetSenderPid(GDBusConnection *connection, const gchar *sender);
  int RegisterDbusInterface(const std::string& port_name);
  void ReceiveMessage(const char* sender_appid,
                      GDBusMethodInvocation* invocation);
  std::string GetInterfaceName(const std::string& target_appid,
                               const std::string& port_name);
  static void OnNameAppeared(GDBusConnection *connection,
                             const gchar *name,
                             const gchar *name_owner,
                             gpointer user_data);
  static void OnNameVanished(GDBusConnection *connection,
                             const gchar *name,
                             gpointer user_data);
  static void OnResultReceived(GObject* source_object,
                               GAsyncResult* res,
                               gpointer user_data);

 private:
  IEventListener* listener_ = nullptr;
  int registration_id_ = 0;
  bool mock_;
  AccessController ac_;
  IEventWatcher* watcher_ = nullptr;
  guint watcher_id_ = 0;
  std::string watch_appid_;
  std::string watch_port_name_;
  GCancellable* cancellable_ = nullptr;
};

}  // namespace internal
}  // namespace rpc_port

#endif  // FDBROKER_INTERNAL_H_
