/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <cstring>

#include "parcel-internal.h"

namespace rpc_port {

Parcel::Parcel(const unsigned char* buf, unsigned int size)
    : data_(buf, buf + size) {}

void Parcel::WriteByte(char b) {
  Write<char>(b);
}

void Parcel::WriteInt16(short i) {
  Write<short>(i);
}

void Parcel::WriteInt32(int i) {
  Write<int>(i);
}

void Parcel::WriteInt64(long long i) {
  Write<long long>(i);
}

void Parcel::WriteFloat(float f) {
  Write<float>(f);
}

void Parcel::WriteDouble(double d) {
  Write<double>(d);
}

void Parcel::WriteString(const char* str) {
  int l = std::strlen(str) + 1;

  Write<int>(l);
  std::copy(str, str + l, std::back_inserter(data_));
}

void Parcel::WriteBool(bool b) {
  Write<bool>(b);
}

void Parcel::WriteBundle(bundle* b) {
  bundle_raw *r;
  int len;
  bundle_encode(b, &r, &len);
  unsigned char* c = reinterpret_cast<unsigned char*>(r);

  Write<int>(len + 1);
  std::copy(c, c + len + 1, std::back_inserter(data_));
  free(r);
}

void Parcel::WriteArrayCount(int count) {
  Write<int>(count);
}

char Parcel::ReadByte() {
  return Read<char>();
}

short Parcel::ReadInt16() {
  return Read<short>();
}

int Parcel::ReadInt32() {
  return Read<int>();
}

long long Parcel::ReadInt64() {
  return Read<long long>();
}

float Parcel::ReadFloat() {
  return Read<float>();
}

double Parcel::ReadDouble() {
  return Read<double>();
}

const char* Parcel::ReadString() {
  int l = Read<int>();

  if (l <= 0)
    return nullptr;
  if (reader_ + l > data_.size())
      return nullptr;
  if (data_[reader_ + l - 1] != 0)
    return nullptr;
  const char* str = reinterpret_cast<const char*>(&data_[reader_]);
  reader_ += l;

  return str;
}

bool Parcel::ReadBool() {
  return Read<bool>();
}

bundle* Parcel::ReadBundle() {
  int l = Read<int>();

  if (l <= 0)
    return nullptr;
  if (reader_ + l > data_.size())
      return nullptr;
  if (data_[reader_ + l - 1] != 0)
    return nullptr;
  const bundle_raw* str = reinterpret_cast<const bundle_raw*>(&data_[reader_]);
  reader_ += l;

  bundle* b = bundle_decode(str, l - 1);

  return b;
}

int Parcel::ReadArrayCount() {
  return Read<int>();
}

const std::vector<unsigned char>& Parcel::GetRaw() {
  return data_;
}

void Parcel::Write(const unsigned char* buf, unsigned int size) {
  std::copy(buf, buf + size, std::back_inserter(data_));
}

void Parcel::Read(unsigned char* buf, unsigned int size) {
  if (reader_ + size > data_.size())
      return;
  std::copy(&data_[reader_], &data_[reader_] + size, buf);
  reader_ += size;
}

}  // namespace rpc_port
