/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>
#include <poll.h>
#include <unistd.h>
#include <uuid/uuid.h>
#include <dlog.h>

#include "rpc-port.h"
#include "port-internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "RPC_PORT"

#define MAX_CNT 100
#define MAX_SLEEP 100
#define MIN_SLEEP 5
#define BASE_SLEEP 1000 * 1000

namespace rpc_port {
namespace internal {

Port::Port(int fd, std::string id)
    : fd_(fd), id_(std::move(id)) {
  char uuid[37];
  uuid_t u;
  uuid_generate(u);
  uuid_unparse(u, uuid);
  instance_ = std::string(uuid) + ":" + id;
}

Port::Port(int fd, std::string id, std::string instance)
    : fd_(fd), id_(std::move(id)), instance_(std::move(instance)) {}

Port::~Port() {
  close(fd_);
}

int Port::Read(void* buf, unsigned int size) {
  unsigned int left = size;
  ssize_t nb;
  struct timespec TRY_SLEEP_TIME = { 0, MIN_SLEEP * BASE_SLEEP};
  int bytes_read = 0;
  char* buffer = static_cast<char*>(buf);
  int max_timeout = MAX_CNT * MAX_SLEEP; /* 10 sec */
  std::lock_guard<std::recursive_mutex> lock(mutex_);

  while (left) {
    nb = read(fd_, buffer, left);
    if (nb == 0) {
      LOGE("read_socket: ...read EOF, socket closed %d: nb %d\n", fd_, nb);
      return RPC_PORT_ERROR_IO_ERROR;
    } else if (nb == -1) {
      if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK) {
        LOGE("read_socket: %d errno, sleep and retry ...", errno);
        nanosleep(&TRY_SLEEP_TIME, 0);
        max_timeout -= (TRY_SLEEP_TIME.tv_nsec / (BASE_SLEEP));
        if (max_timeout <= 0) {
          LOGE("read_socket: ...timed out fd %d: errno %d", fd_, errno);
          return RPC_PORT_ERROR_IO_ERROR;
        }
        TRY_SLEEP_TIME.tv_nsec *= 2;
        if (TRY_SLEEP_TIME.tv_nsec > (MAX_SLEEP * BASE_SLEEP))
          TRY_SLEEP_TIME.tv_nsec = MAX_SLEEP * BASE_SLEEP;
        continue;
      }

      LOGE("read_socket: ...error fd %d: errno %d\n", fd_, errno);
      return RPC_PORT_ERROR_IO_ERROR;
    }

    left -= nb;
    buffer += nb;
    bytes_read += nb;
    TRY_SLEEP_TIME.tv_nsec = MIN_SLEEP * BASE_SLEEP;
  }

  return RPC_PORT_ERROR_NONE;
}

int Port::Write(const void* buf, unsigned int size) {
  unsigned int left = size;
  ssize_t nb;
  struct timespec TRY_SLEEP_TIME = { 0, MIN_SLEEP * BASE_SLEEP };
  struct pollfd fds[1];
  int ret;
  int bytes_write = 0;
  const char* buffer = static_cast<const char*>(buf);
  int max_timeout = MAX_CNT * MAX_SLEEP; /* 10 sec */
  struct timespec start_time = { 0, };
  struct timespec end_time = { 0, };
  std::lock_guard<std::recursive_mutex> lock(mutex_);

  fds[0].fd = fd_;
  fds[0].events = POLLOUT;
  fds[0].revents = 0;

  clock_gettime(CLOCK_MONOTONIC, &start_time);
  ret = poll(fds, 1, MAX_SLEEP * MAX_CNT);
  clock_gettime(CLOCK_MONOTONIC, &end_time);
  if (ret == 0) {
    LOGE("write_socket: : fd %d poll timeout", fd_);
    return RPC_PORT_ERROR_IO_ERROR;
  }

  max_timeout -= (((end_time.tv_sec - start_time.tv_sec) * 1000) +
      ((end_time.tv_nsec - start_time.tv_nsec) / (BASE_SLEEP)));
  if (max_timeout <= 0) {
    LOGE("write_socket: ...timed out fd %d: errno %d", fd_, errno);
    return RPC_PORT_ERROR_IO_ERROR;
  }

  while (left) {
    nb = send(fd_, buffer, left, MSG_NOSIGNAL);
    if (nb == -1) {
      if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK) {
        LOGE("write_socket: %d errno, sleep and retry ...", errno);
        nanosleep(&TRY_SLEEP_TIME, 0);
        max_timeout -= (TRY_SLEEP_TIME.tv_nsec / (BASE_SLEEP));
        if (max_timeout <= 0) {
          LOGE("write_socket: ...timed out fd %d: errno %d", fd_, errno);
          return RPC_PORT_ERROR_IO_ERROR;
        }
        TRY_SLEEP_TIME.tv_nsec *= 2;
        if (TRY_SLEEP_TIME.tv_nsec > (MAX_SLEEP * BASE_SLEEP))
          TRY_SLEEP_TIME.tv_nsec = MAX_SLEEP * BASE_SLEEP;
        continue;
      }

      LOGE("write_socket: ...error fd %d: errno %d\n", fd_, errno);
      return RPC_PORT_ERROR_IO_ERROR;
    }

    left -= nb;
    buffer += nb;
    bytes_write += nb;
    TRY_SLEEP_TIME.tv_nsec = MIN_SLEEP * BASE_SLEEP;
  }

  return RPC_PORT_ERROR_NONE;
}

}  // namespace internal
}  // namespace rpc_port
