/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <dlog.h>

#include "proxy-internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "RPC_PORT"

#define EILLEGALACCESS 127

namespace rpc_port {
namespace internal {

Proxy::Proxy(bool mock)
    : fd_broker_(mock) {}

Proxy::~Proxy() {
  LOGD("Proxy::~Proxy");
  if (conn_timer_) {
    g_source_remove(conn_timer_);
    conn_timer_ = 0;
  }
}

gboolean Proxy::OnSocketDisconnected(GIOChannel *gio, GIOCondition cond,
                                     gpointer data) {
  Proxy* proxy = static_cast<Proxy*>(data);
  IEventListener* listener = proxy->listener_;
  int fd = g_io_channel_unix_get_fd(gio);

  LOGW("Socket was disconnected. fd(%d)", fd);

  if (proxy->main_port_.get()->GetFd() == fd) {
    proxy->listener_ = nullptr;
    proxy->main_port_.get()->SetDisconnectedSource(0);
    listener->OnDisconnected(proxy->target_appid_);
  } else if (proxy->delegate_port_.get()->GetFd() == fd) {
    proxy->delegate_port_.get()->SetDisconnectedSource(0);
  }

  return FALSE;
}

gboolean Proxy::OnDataReceived(GIOChannel *gio, GIOCondition cond,
                               gpointer data) {
  Proxy* proxy = static_cast<Proxy*>(data);
  int fd = g_io_channel_unix_get_fd(gio);
  char buffer[4];

  if (proxy->delegate_port_.get()->GetFd() == fd) {
    if (recv(fd, buffer, sizeof(buffer), MSG_PEEK | MSG_DONTWAIT) == 0) {
      LOGW("Socket was disconnected by stub. fd(%d)", fd);
      IEventListener* listener = proxy->listener_;
      proxy->listener_ = nullptr;
      proxy->delegate_port_.get()->SetSource(0);
      if (listener)
        listener->OnDisconnected(proxy->target_appid_);
      return FALSE;
    }

    proxy->listener_->OnReceived(proxy->target_appid_);
  }

  return TRUE;
}

void Proxy::OnPortRejected(const std::string& appid) {
  if (listener_ == nullptr)
    return;

  IEventListener* listener = listener_;
  listener_ = nullptr;
  listener->OnRejected(appid);
}

void Proxy::OnPortAppeared(const std::string& appid,
                           const std::string& port_name) {
  LOGD("endpoint(%s), port_name(%s)", appid.c_str(), port_name.c_str());
  if (conn_timer_) {
    g_source_remove(conn_timer_);
    conn_timer_ = 0;
  }

  if (listener_ == nullptr)
    return;

  fds_[0] = 0;
  fds_[1] = 0;
  int r = fd_broker_.Send(appid, port_name, &fds_);
  if (r <= 0) {
    IEventListener* listener = listener_;
    listener_ = nullptr;
    if (r == -EILLEGALACCESS)
      listener->OnRejected(appid);
    else
      listener->OnDisconnected(appid);
    return;
  }

  LOGW("[__OnPortAppeared__] fds[0]: %d, fds[1]: %d", fds_[0], fds_[1]);
}

void Proxy::OnPortVanished(const std::string& appid,
                           const std::string& port_name) {
  LOGW("[__OnPortVanished__] endpoint(%s), port_name(%s)",
      appid.c_str(), port_name.c_str());
}

void Proxy::OnPortConnected(const std::string& appid,
                            const std::string& port_name) {
  LOGW("[__OnPortConnected__] endpoint(%s), port_name(%s)",
      appid.c_str(), port_name.c_str());

  main_port_.reset(new ProxyPort(this, fds_[0], appid, false));
  delegate_port_.reset(new ProxyPort(this, fds_[1], appid));
  listener_->OnConnected(appid, main_port_.get());
}

void Proxy::OnPortDisconnected(const std::string& appid,
                               const std::string& port_name, bool cancel) {
  LOGW("[__OnPortDisconnected__] endporint(%s), port_name(%s)",
      appid.c_str(), port_name.c_str());

  if (cancel) {
    close(fds_[0]);
    close(fds_[1]);
  }

  IEventListener* listener = listener_;
  listener_ = nullptr;
  listener->OnDisconnected(appid);
}

int Proxy::Connect(std::string appid, std::string port_name,
                   IEventListener* ev) {
  if (ev == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  if (listener_ != nullptr) {
    LOGD("Already connected");
    return RPC_PORT_ERROR_INVALID_PARAMETER;
  }

  listener_ = ev;
  target_appid_ = std::move(appid);
  port_name_ = std::move(port_name);
  int r = fd_broker_.Watch(this, target_appid_, port_name_);
  if (r < 0) {
    listener_ = nullptr;
    if (r == -EILLEGALACCESS)
      return RPC_PORT_ERROR_PERMISSION_DENIED;

    return RPC_PORT_ERROR_IO_ERROR;
  }

  if (conn_timer_)
    g_source_remove(conn_timer_);
  conn_timer_ = g_timeout_add(10 * 1000, DbusNameTimeout, this);

  return RPC_PORT_ERROR_NONE;
}

gboolean Proxy::DbusNameTimeout(gpointer user_data) {
  Proxy* obj = static_cast<Proxy*>(user_data);

  LOGW("[__DbusNameTimeout__] endpoint(%s)", obj->target_appid_.c_str());
  if (obj->listener_)
    obj->listener_->OnRejected(obj->target_appid_);

  obj->conn_timer_ = 0;
  return G_SOURCE_REMOVE;
}

Proxy::ProxyPort::ProxyPort(Proxy* parent, int fd, const std::string& id,
    bool receive) : Port(fd, id), parent_(parent) {
  Watch(receive);
}

Proxy::ProxyPort::~ProxyPort() {
  if (disconn_src_ > 0)
    g_source_remove(disconn_src_);

  if (src_ > 0)
    g_source_remove(src_);

  if (gioc_ != nullptr)
    g_io_channel_unref(gioc_);
}

int Proxy::ProxyPort::Watch(bool receive) {
  char buf[1024];
  int fd = GetFd();

  gioc_ = g_io_channel_unix_new(fd);
  if (!gioc_) {
    LOGE("Error is %s", strerror_r(errno, buf, sizeof(buf)));
    return -1;
  }

  disconn_src_ = g_io_add_watch(gioc_,
      (GIOCondition)(G_IO_ERR | G_IO_HUP | G_IO_NVAL),
      Proxy::OnSocketDisconnected, parent_);
  if (disconn_src_ == 0) {
    LOGE("Failed to add watch on socket");
    g_io_channel_unref(gioc_);
    gioc_ = nullptr;
    return -1;
  }

  if (!receive)
    return 0;

  src_ = g_io_add_watch(gioc_,
      (GIOCondition)(G_IO_IN),
      Proxy::OnDataReceived, parent_);
  if (src_ == 0) {
    LOGE("Failed to add watch on socket");
    g_source_remove(disconn_src_);
    disconn_src_ = 0;
    g_io_channel_unref(gioc_);
    gioc_ = nullptr;
    return -1;
  }

  return 0;
}

void Proxy::ProxyPort::SetDisconnectedSource(int sourceId) {
  disconn_src_ = sourceId;
}

void Proxy::ProxyPort::SetSource(int sourceId) {
  src_ = sourceId;
}

}  // namespace internal
}  // namespace rpc_port
