/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef PROXY_INTERNAL_H_
#define PROXY_INTERNAL_H_

#include <glib.h>
#include <gio/gio.h>
#include <glib-unix.h>

#include <string>
#include <memory>

#include "port-internal.h"
#include "fdbroker-internal.h"

namespace rpc_port {
namespace internal {

class Proxy : public FdBroker::IEventWatcher {
 public:
  explicit Proxy(bool mock = false);
  virtual ~Proxy();

  class IEventListener {
   public:
    virtual void OnConnected(const std::string& endpoint, Port* port) = 0;
    virtual void OnDisconnected(const std::string& endpoint) = 0;
    virtual void OnRejected(const std::string& endpoint) = 0;
    virtual void OnReceived(const std::string& endpoint) = 0;
  };

  int Connect(std::string appid, std::string port_name, IEventListener* ev);

  std::shared_ptr<Port> GetPort() const {
    return main_port_;
  }

  std::shared_ptr<Port> GetDelegatePort() const {
    return delegate_port_;
  }

  const std::string& GetPortName() {
    return port_name_;
  }

 private:
  class ProxyPort : public Port {
   public:
    ProxyPort(Proxy* parent, int fd, const std::string& id, bool receive = true);
    virtual ~ProxyPort();
    void SetDisconnectedSource(int sourceId);
    void SetSource(int sourceId);

   private:
    int Watch(bool receive);

   private:
    GIOChannel* gioc_ = nullptr;
    int disconn_src_ = 0;
    int src_ = 0;
    Proxy* parent_ = nullptr;
  };

 private:
  static gboolean OnSocketDisconnected(GIOChannel *gio, GIOCondition cond,
                                       gpointer data);
  static gboolean OnDataReceived(GIOChannel *gio, GIOCondition cond,
                                 gpointer data);
  static gboolean DbusNameTimeout(gpointer user_data);
  void OnPortAppeared(const std::string& appid,
                      const std::string& port_name) override;
  void OnPortVanished(const std::string& appid,
                      const std::string& port_name) override;
  void OnPortRejected(const std::string& appid) override;
  void OnPortConnected(const std::string& appid,
                       const std::string& port_name) override;
  void OnPortDisconnected(const std::string& appid,
                          const std::string& port_name,
                          bool cancel = false) override;

 private:
  std::string port_name_;
  std::shared_ptr<ProxyPort> main_port_;
  std::shared_ptr<ProxyPort> delegate_port_;
  IEventListener* listener_ = nullptr;
  FdBroker fd_broker_;
  std::string target_appid_;
  int fds_[2];
  guint conn_timer_ = 0;
};

}  // namespace internal
}  // namespace rpc_port

#endif  // PROXY_INTERNAL_H_
