/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <string.h>

#include "rpc-port-parcel.h"

#include "parcel-internal.h"
#include "port-internal.h"

#define MAX_PARCEL_SIZE   (1024 * 1024 * 10)

#undef RPC_API
#define RPC_API extern "C" __attribute__((visibility("default")))

using namespace rpc_port;

RPC_API int rpc_port_parcel_create(rpc_port_parcel_h* h) {
  Parcel* p = new Parcel();

  *h = p;

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_create_from_port(rpc_port_parcel_h* h,
    rpc_port_h port) {
  int len;
  unsigned char* buf;

  if (port == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  internal::Port* pt = static_cast<internal::Port*>(port);
  {
    std::lock_guard<std::recursive_mutex> lock(pt->GetMutex());
    int ret = rpc_port_read(port, &len, 4);
    if (ret != 0)
      return ret;

    if (len <= 0 || len > MAX_PARCEL_SIZE)
      return RPC_PORT_ERROR_IO_ERROR;
    buf = new unsigned char[len];
    ret = rpc_port_read(port, buf, len);
    if (ret != 0) {
      delete[] buf;
      return ret;
    }
  }

  Parcel* p = new Parcel(buf, len);
  delete[] buf;

  *h = p;

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_send(rpc_port_parcel_h h, rpc_port_h port) {
  if (h == nullptr || port == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);
  int len = p->GetRaw().size();

  if (len <= 0)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  internal::Port* pt = static_cast<internal::Port*>(port);
  {
    std::lock_guard<std::recursive_mutex> lock(pt->GetMutex());
    int ret = rpc_port_write(port, &len, 4);
    if (ret != 0)
      return ret;

    ret = rpc_port_write(port, &*(p->GetRaw().cbegin()), len);
    if (ret != 0)
      return ret;
  }

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_destroy(rpc_port_parcel_h h) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  delete static_cast<Parcel*>(h);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_byte(rpc_port_parcel_h h, char b) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteByte(b);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_int16(rpc_port_parcel_h h, short i) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteInt16(i);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_int32(rpc_port_parcel_h h, int i) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteInt32(i);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_int64(rpc_port_parcel_h h, long long i) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteInt64(i);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_float(rpc_port_parcel_h h, float f) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteFloat(f);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_double(rpc_port_parcel_h h, double d) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteDouble(d);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_string(rpc_port_parcel_h h, const char* str) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteString(str);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_bool(rpc_port_parcel_h h, bool b) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteBool(b);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_bundle(rpc_port_parcel_h h, bundle* b) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteBundle(b);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write_array_count(rpc_port_parcel_h h, int count) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->WriteArrayCount(count);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_write(rpc_port_parcel_h h,
    rpc_port_parcelable_t* parcelable, void* data) {
  if (parcelable == nullptr || parcelable->to == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  parcelable->to(h, data);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_byte(rpc_port_parcel_h h, char* b) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *b = p->ReadByte();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_int16(rpc_port_parcel_h h, short* i) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *i = p->ReadInt16();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_int32(rpc_port_parcel_h h, int* i) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *i = p->ReadInt32();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_int64(rpc_port_parcel_h h, long long* i) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *i = p->ReadInt64();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_float(rpc_port_parcel_h h, float* f) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *f = p->ReadFloat();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_double(rpc_port_parcel_h h, double* d) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *d = p->ReadDouble();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_string(rpc_port_parcel_h h, char** str) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);
  const char* read = p->ReadString();
  *str = read == nullptr ? strdup("") : strdup(read);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_bool(rpc_port_parcel_h h, bool* b) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *b = p->ReadBool();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_bundle(rpc_port_parcel_h h, bundle** b) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);
  bundle* read = p->ReadBundle();

  *b = read == nullptr ? bundle_create() : read;

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read_array_count(rpc_port_parcel_h h, int* count) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  *count = p->ReadArrayCount();

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_read(rpc_port_parcel_h h,
    rpc_port_parcelable_t* parcelable, void* data) {
  if (parcelable == nullptr || parcelable->from == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  parcelable->from(h, data);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_burst_read(rpc_port_parcel_h h, unsigned char *buf,
    unsigned int size) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->Read(buf, size);

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_parcel_burst_write(rpc_port_parcel_h h,
    const unsigned char *buf, unsigned int size) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  Parcel* p = static_cast<Parcel*>(h);

  p->Write(buf, size);

  return RPC_PORT_ERROR_NONE;
}
