/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>

#include <thread>
#include <mutex>

#include "rpc-port.h"
#include "port-internal.h"
#include "proxy-internal.h"
#include "stub-internal.h"

#undef RPC_API
#define RPC_API extern "C" __attribute__((visibility("default")))

namespace {
using namespace rpc_port::internal;

template<typename T>
class Event {
 public:
  Event(T cb, void* user_data)
      : cb_(cb), user_data_(user_data) {}

  T cb_;
  void* user_data_;
};

class ProxyExt : public Proxy, public Proxy::IEventListener {
 public:
  explicit ProxyExt(bool mock = false) : Proxy(mock) {}
  virtual ~ProxyExt() = default;

  void AddConnectedEventListener(rpc_port_proxy_connected_event_cb cb,
                            void* user_data) {
    connected_events_.emplace_back(
        new Event<rpc_port_proxy_connected_event_cb>(cb, user_data));
  }

  void AddDisconnectedEventListener(rpc_port_proxy_disconnected_event_cb cb,
                                    void* user_data) {
    disconnected_events_.emplace_back(
        new Event<rpc_port_proxy_disconnected_event_cb>(cb, user_data));
  }

  void AddRejectedEventListener(rpc_port_proxy_rejected_event_cb cb,
                                void* user_data) {
    rejected_events_.emplace_back(
        new Event<rpc_port_proxy_rejected_event_cb>(cb, user_data));
  }

  void AddReceivedEventListener(rpc_port_proxy_received_event_cb cb,
                                void* user_data) {
    received_events_.emplace_back(
        new Event<rpc_port_proxy_received_event_cb>(cb, user_data));
  }

  void OnConnected(const std::string& endpoint, Port* port) override {
    for (auto& ev : connected_events_) {
      ev->cb_(endpoint.c_str(), GetPortName().c_str(), port,
              ev->user_data_);
    }
  }

  void OnDisconnected(const std::string& endpoint) override {
    for (auto& ev : disconnected_events_) {
      ev->cb_(endpoint.c_str(), GetPortName().c_str(), ev->user_data_);
    }
  }

  void OnRejected(const std::string& endpoint) override {
    for (auto& ev : rejected_events_) {
      ev->cb_(endpoint.c_str(), GetPortName().c_str(), ev->user_data_);
    }
  }

  void OnReceived(const std::string& endpoint) override {
    for (auto& ev : received_events_) {
      ev->cb_(endpoint.c_str(), GetPortName().c_str(), ev->user_data_);
    }
  }

  std::recursive_mutex& GetMutex() const {
    return mutex_;
  }

 private:
  std::list<std::unique_ptr<Event<rpc_port_proxy_connected_event_cb>>>
      connected_events_;
  std::list<std::unique_ptr<Event<rpc_port_proxy_disconnected_event_cb>>>
      disconnected_events_;
  std::list<std::unique_ptr<Event<rpc_port_proxy_rejected_event_cb>>>
      rejected_events_;
  std::list<std::unique_ptr<Event<rpc_port_proxy_received_event_cb>>>
      received_events_;
  mutable std::recursive_mutex mutex_;
};

class StubExt : public Stub, public Stub::IEventListener {
 public:
  explicit StubExt(const std::string& port, bool mock = false)
      : Stub(port, mock) {}
  virtual ~StubExt() = default;

  void AddConnectedEventListener(rpc_port_stub_connected_event_cb cb,
                                 void* user_data) {
    connected_events_.emplace_back(
        new Event<rpc_port_stub_connected_event_cb>(cb, user_data));
  }

  void AddDisconnectedEventListener(rpc_port_stub_disconnected_event_cb cb,
                                    void* user_data) {
    disconnected_events_.emplace_back(
        new Event<rpc_port_stub_disconnected_event_cb>(cb, user_data));
  }

  void AddReceivedEventListener(rpc_port_stub_received_event_cb cb,
                                void* user_data) {
    received_events_.emplace_back(
        new Event<rpc_port_stub_received_event_cb>(cb, user_data));
  }

  void OnConnected(const std::string& sender,
                   const std::string& instance) override {
    for (auto& ev : connected_events_) {
      ev->cb_(sender.c_str(), instance.c_str(), ev->user_data_);
    }
  }

  void OnDisconnected(const std::string& sender,
                      const std::string& instance) override {
    for (auto& ev : disconnected_events_) {
      ev->cb_(sender.c_str(), instance.c_str(), ev->user_data_);
    }
  }

  int OnReceived(const std::string& sender,
                 const std::string& instance, Port* port) override {
    for (auto& ev : received_events_) {
      int ret = ev->cb_(sender.c_str(), instance.c_str(), port,
          ev->user_data_);
      if (ret != 0)
        return -1;
    }

    return 0;
  }

  std::recursive_mutex& GetMutex() const {
    return mutex_;
  }

 private:
  std::list<std::unique_ptr<Event<rpc_port_stub_connected_event_cb>>>
      connected_events_;
  std::list<std::unique_ptr<Event<rpc_port_stub_disconnected_event_cb>>>
      disconnected_events_;
  std::list<std::unique_ptr<Event<rpc_port_stub_received_event_cb>>>
      received_events_;
  mutable std::recursive_mutex mutex_;
};

}  // namespace

RPC_API int rpc_port_read(rpc_port_h h, void* buf, unsigned int size) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto port = static_cast<Port*>(h);

  return port->Read(buf, size);
}

RPC_API int rpc_port_write(rpc_port_h h, const void* buf, unsigned int size) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto port = static_cast<Port*>(h);

  return port->Write(buf, size);
}

RPC_API int rpc_port_proxy_create(rpc_port_proxy_h* h) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = new ::ProxyExt();

  *h = p;
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_proxy_create_mockup(rpc_port_proxy_h* h) {
  auto p = new ::ProxyExt(true);

  *h = p;
  return 0;
}

RPC_API int rpc_port_proxy_destroy(rpc_port_proxy_h h) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);

  delete p;
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_proxy_connect(rpc_port_proxy_h h, const char* appid,
    const char* port) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  return p->Connect(appid, port, p);
}

RPC_API int rpc_port_proxy_add_connected_event_cb(rpc_port_proxy_h h,
    rpc_port_proxy_connected_event_cb cb, void *user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddConnectedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_proxy_add_disconnected_event_cb(rpc_port_proxy_h h,
    rpc_port_proxy_disconnected_event_cb cb, void* user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddDisconnectedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_proxy_add_rejected_event_cb(rpc_port_proxy_h h,
    rpc_port_proxy_rejected_event_cb cb, void* user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddRejectedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_proxy_add_received_event_cb(rpc_port_proxy_h h,
    rpc_port_proxy_received_event_cb cb, void* user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddReceivedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_proxy_get_port(rpc_port_proxy_h h,
    rpc_port_port_type_e type, rpc_port_h* port) {
  if (h == nullptr || port == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::ProxyExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());
  rpc_port_h ret_port;

  switch (type) {
    case RPC_PORT_PORT_MAIN:
      ret_port = static_cast<rpc_port_h>(p->GetPort().get());
      if (ret_port == nullptr)
        return RPC_PORT_ERROR_IO_ERROR;
      *port = ret_port;
      break;
    case RPC_PORT_PORT_CALLBACK:
      ret_port = static_cast<rpc_port_h>(p->GetDelegatePort().get());
      if (ret_port == nullptr)
        return RPC_PORT_ERROR_IO_ERROR;
      *port = ret_port;
      break;
  }

  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_create(rpc_port_stub_h* h, const char* port_name) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = new ::StubExt(port_name);

  *h = p;
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_create_mockup(rpc_port_stub_h* h,
                                        const char* port_name) {
  if (h == nullptr)
    return -1;

  auto p = new ::StubExt(port_name, true);

  *h = p;
  return 0;
}

RPC_API int rpc_port_stub_destroy(rpc_port_stub_h h) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);

  delete p;
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_listen(rpc_port_stub_h h) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  return p->Listen(p);
}

RPC_API int rpc_port_stub_add_privilege(rpc_port_stub_h h,
    const char* privilege) {
  if (h == nullptr || privilege == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddPrivilege(privilege);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_set_trusted(rpc_port_stub_h h,
    const bool trusted) {
  if (h == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->SetTrusted(trusted);
  return RPC_PORT_ERROR_NONE;
}


RPC_API int rpc_port_stub_add_connected_event_cb(rpc_port_stub_h h,
    rpc_port_stub_connected_event_cb cb, void* user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddConnectedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_add_disconnected_event_cb(rpc_port_stub_h h,
    rpc_port_stub_disconnected_event_cb cb, void* user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddDisconnectedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_add_received_event_cb(rpc_port_stub_h h,
    rpc_port_stub_received_event_cb cb, void* user_data) {
  if (h == nullptr || cb == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());

  p->AddReceivedEventListener(cb, user_data);
  return RPC_PORT_ERROR_NONE;
}

RPC_API int rpc_port_stub_get_port(rpc_port_stub_h h,
    rpc_port_port_type_e type, const char *instance, rpc_port_h* port) {
  if (h == nullptr || port == nullptr || instance == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  auto p = static_cast<::StubExt*>(h);
  std::lock_guard<std::recursive_mutex> lock(p->GetMutex());
  rpc_port_h ret_port;

  switch (type) {
    case RPC_PORT_PORT_MAIN:
      ret_port = static_cast<rpc_port_h>(p->FindPort(instance).get());
      if (ret_port == nullptr)
        return RPC_PORT_ERROR_IO_ERROR;
      *port = ret_port;
      break;
    case RPC_PORT_PORT_CALLBACK:
      ret_port = static_cast<rpc_port_h>(p->FindDelegatePort(instance).get());
      if (ret_port == nullptr)
        return RPC_PORT_ERROR_IO_ERROR;
      *port = ret_port;
      break;
  }

  return RPC_PORT_ERROR_NONE;
}

