/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <dlog.h>
#include <aul.h>
#include <aul_rpc_port.h>

#include "stub-internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "RPC_PORT"

namespace rpc_port {
namespace internal {

Stub::Stub(const std::string& port_name, bool mock)
    : fd_broker_(mock), port_name_(port_name) {}

Stub::~Stub() {
  LOGD("Stub::~Stub");
}

int Stub::Listen(IEventListener* ev) {
  if (ev == nullptr)
    return RPC_PORT_ERROR_INVALID_PARAMETER;

  listener_ = ev;
  return fd_broker_.Listen(this, port_name_);
}

void Stub::AddPrivilege(const std::string& privilege) {
  AccessController& ac = fd_broker_.GetAccessController();
  ac.AddPrivilege(privilege);
}

void Stub::SetTrusted(const bool trusted) {
  AccessController& ac = fd_broker_.GetAccessController();
  ac.SetTrusted(trusted);
}

std::shared_ptr<Port> Stub::FindPort(const std::string& instance) const {
  for (auto& p : ports_) {
    if (p->GetInstance() == instance && !p->IsDelegate()) {
      return p;
    }
  }

  return {};
}

std::shared_ptr<Port> Stub::FindDelegatePort(
    const std::string& instance) const {
  for (auto& p : ports_) {
    if (p->GetInstance() == instance && p->IsDelegate()) {
      return p;
    }
  }

  return {};
}

void Stub::RemoveAcceptedPorts(std::string instance) {
  auto iter = ports_.begin();
  while (iter != ports_.end()) {
    if ((*iter)->GetInstance().compare(instance) == 0) {
      LOGI("Close: fd(%d)", (*iter)->GetFd());
      iter = ports_.erase(iter);
    } else {
      iter++;
    }
  }
}

gboolean Stub::OnDataReceived(GIOChannel *gio, GIOCondition cond,
                              gpointer data) {
  Stub* stub = static_cast<Stub*>(data);
  int fd = g_io_channel_unix_get_fd(gio);
  char buffer[4];

  for (auto& p : stub->ports_) {
    if (p->GetFd() == fd && !p->IsDelegate()) {
      if (recv(fd, buffer, sizeof(buffer), MSG_PEEK | MSG_DONTWAIT) == 0) {
        LOGW("Socket was disconnected from proxy. fd(%d)", fd);
        stub->listener_->OnDisconnected(p->GetId(), p->GetInstance());
        stub->RemoveAcceptedPorts(p->GetInstance());

        if (aul_rpc_port_notify_rpc_finished() != AUL_R_OK)
          LOGW("Failed to notify rpc finished");

        return FALSE;
      }

      int ret = stub->listener_->OnReceived(p->GetId(), p->GetInstance(),
                                            p.get());

      if (ret != 0) {
        LOGW("Invalid protocol");
        stub->listener_->OnDisconnected(p->GetId(), p->GetInstance());
        stub->RemoveAcceptedPorts(p->GetInstance());

        if (aul_rpc_port_notify_rpc_finished() != AUL_R_OK)
          LOGW("Failed to notify rpc finished");

        return FALSE;
      }

      break;
    }
  }

  return TRUE;
}

gboolean Stub::OnSocketDisconnected(GIOChannel *gio, GIOCondition cond,
                                    gpointer data) {
  Stub* stub = static_cast<Stub*>(data);
  int fd = g_io_channel_unix_get_fd(gio);

  LOGW("Socket was disconnected. fd(%d)", fd);
  for (auto& p : stub->ports_) {
    if (p->GetFd() == fd) {
      stub->listener_->OnDisconnected(p->GetId(), p->GetInstance());
      if (aul_rpc_port_notify_rpc_finished() != AUL_R_OK)
        LOGW("Failed to notify rpc finished");
      stub->RemoveAcceptedPorts(p->GetInstance());
      break;
    }
  }

  return FALSE;
}

void Stub::OnFdReceived(const std::string& sender, int fds[2]) {
  LOGW("[__OnFdReceived__] fds[0]: %d, fds[1]: %d", fds[0], fds[1]);
  auto* main_port = new AcceptedPort(this, false, fds[0], sender, true);
  ports_.emplace_back(main_port);
  ports_.emplace_back(new AcceptedPort(this, true, fds[1], sender,
      main_port->GetInstance(), false));
  for (auto& p : ports_) {
    if (p->GetFd() == fds[0]) {
      listener_->OnConnected(p->GetId(), p->GetInstance());
      break;
    }
  }
}

Stub::AcceptedPort::AcceptedPort(Stub* parent, bool isDelegate, int fd,
    std::string id, std::string inst, bool watch)
    : Port(fd, std::move(id), std::move(inst)), parent_(parent),
      is_delegate_(isDelegate) {
  Watch(watch);
}

Stub::AcceptedPort::AcceptedPort(Stub* parent, bool isDelegate, int fd,
    std::string id, bool watch)
    : Port(fd, std::move(id)), parent_(parent),
      is_delegate_(isDelegate) {
  Watch(watch);
}

Stub::AcceptedPort::~AcceptedPort() {
  if (disconn_src_ > 0)
    g_source_remove(disconn_src_);

  if (src_ > 0)
    g_source_remove(src_);

  if (gioc_ != nullptr)
    g_io_channel_unref(gioc_);
}

int Stub::AcceptedPort::Watch(bool receive) {
  char buf[1024];
  int fd = GetFd();

  gioc_ = g_io_channel_unix_new(fd);
  if (!gioc_) {
    LOGE("Error is %s", strerror_r(errno, buf, sizeof(buf)));
    return -1;
  }

  disconn_src_ = g_io_add_watch(gioc_,
                                (GIOCondition)(G_IO_ERR | G_IO_HUP | G_IO_NVAL),
                                Stub::OnSocketDisconnected, parent_);
  if (disconn_src_ == 0) {
    LOGE("fail to add watch on socket");
    g_io_channel_unref(gioc_);
    gioc_ = nullptr;
    return -1;
  }

  if (!receive)
    return 0;

  src_ = g_io_add_watch(gioc_,
                        (GIOCondition)(G_IO_IN),
                        Stub::OnDataReceived, parent_);
  if (src_ == 0) {
    LOGE("fail to add watch on socket");
    g_source_remove(disconn_src_);
    disconn_src_ = 0;
    g_io_channel_unref(gioc_);
    gioc_ = nullptr;
    return -1;
  }

  return 0;
}

}  // namespace internal
}  // namespace rpc_port
