/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <iostream>
#include <stdbool.h>
#include <stdexcept>
#include <glib.h>

#include "rpc-port-internal.h"

using ::testing::AtLeast;

class RpcPortBase : public ::testing::Test {
 public:
  virtual void SetUp() {
    mainloop_ = g_main_loop_new(nullptr, FALSE);

    int ret = rpc_port_proxy_create_mockup(&proxy_handle_);

    ASSERT_NE(proxy_handle_, nullptr);
    ASSERT_EQ(ret, 0);

    ret = rpc_port_stub_create_mockup(&stub_handle_, "test_port");
    ASSERT_NE(stub_handle_, nullptr);
    ASSERT_EQ(ret, 0);
  }

  virtual void TearDown() {
    if (proxy_handle_) {
      int ret = rpc_port_proxy_destroy(proxy_handle_);
      ASSERT_EQ(ret, 0);
    }

    if (stub_handle_) {
      int ret = rpc_port_stub_destroy(stub_handle_);
      ASSERT_EQ(ret, 0);
    }

    g_main_loop_unref(mainloop_);
    mainloop_ = nullptr;
  }

  void RunMainLoop() {
    g_main_loop_run(mainloop_);
  }

  void Finish() {
    g_main_loop_quit(mainloop_);
  }

  void KillStub() {
    int ret = rpc_port_stub_destroy(stub_handle_);
    ASSERT_EQ(ret, 0);
    stub_handle_ = nullptr;
  }

  void KillProxy() {
    int ret = rpc_port_proxy_destroy(proxy_handle_);
    ASSERT_EQ(ret, 0);
    proxy_handle_ = nullptr;
  }

  rpc_port_proxy_h proxy_handle_;
  rpc_port_stub_h stub_handle_;
  bool touch_proxy_connected_event_cb_ = false;
  bool touch_stub_connected_event_cb_ = false;
  bool touch_proxy_rejected_event_cb_ = false;

 private:
  static GMainLoop* mainloop_;
};

GMainLoop* RpcPortBase::mainloop_ = nullptr;

class RpcPortConnection : public RpcPortBase {
 public:
  virtual void SetUp() {
    RpcPortBase::SetUp();
    StubSetup();
    ProxySetup();
  }

  virtual void TearDown() {
    RpcPortBase::TearDown();
  }

  void StubSetup() {
    int ret = rpc_port_stub_add_received_event_cb(stub_handle_,
        [](const char* sender, const char* instance,
           rpc_port_h port, void *data) -> int {
          RpcPortConnection* p = static_cast<RpcPortConnection*>(data);
          p->stub_port_ = port;
          rpc_port_stub_get_port(p->stub_handle_, RPC_PORT_PORT_CALLBACK,
              instance, &p->stub_callback_port_);

          p->Finish();
          return 0;
        }, this);
    ASSERT_EQ(ret, 0);

    ret = rpc_port_stub_add_disconnected_event_cb(stub_handle_,
        [](const char* sender, const char* instance, void *data) {
          RpcPortConnection* p = static_cast<RpcPortConnection*>(data);
          p->touch_stub_disconnected_event_cb_ = true;
          p->Finish();
        }, this);
    ASSERT_EQ(ret, 0);

    ret = rpc_port_stub_add_privilege(stub_handle_,
        "http://tizen.org/privilege/appmanager.launch");
    ASSERT_EQ(ret, 0);

    ret = rpc_port_stub_listen(stub_handle_);
    ASSERT_EQ(ret, 0);
  }

  void ProxySetup() {
    int ret = rpc_port_proxy_add_connected_event_cb(proxy_handle_,
        [](const char *ep, const char *port_name, rpc_port_h port, void *data) {
          RpcPortConnection* p = static_cast<RpcPortConnection*>(data);
          p->proxy_port_ = port;
          rpc_port_proxy_get_port(p->proxy_handle_, RPC_PORT_PORT_CALLBACK,
              &p->proxy_callback_port_);
        }, this);
    ASSERT_EQ(ret, 0);

    ret = rpc_port_proxy_add_disconnected_event_cb(proxy_handle_,
        [](const char *ep, const char *port_name, void *data) {
          RpcPortConnection* p = static_cast<RpcPortConnection*>(data);
          p->touch_proxy_disconnected_event_cb_ = true;
          p->Finish();
        }, this);
    ASSERT_EQ(ret, 0);

    ret = rpc_port_proxy_add_received_event_cb(proxy_handle_,
        [](const char *ep, const char *port_name, void *data) {
          RpcPortConnection* p = static_cast<RpcPortConnection*>(data);
          p->touch_proxy_received_event_cb_ = true;
          p->Finish();
        }, this);
    ASSERT_EQ(ret, 0);

    ret = rpc_port_proxy_connect(proxy_handle_, "TestApp", "test_port");
    ASSERT_EQ(ret, 0);
  }

  rpc_port_h proxy_port_ = nullptr;
  rpc_port_h proxy_callback_port_ = nullptr;
  rpc_port_h stub_port_ = nullptr;
  rpc_port_h stub_callback_port_ = nullptr;
  bool touch_proxy_disconnected_event_cb_ = false;
  bool touch_stub_disconnected_event_cb_ = false;
  bool touch_proxy_received_event_cb_ = false;
};

TEST_F(RpcPortBase, rpc_port_event_connect) {
  int ret = rpc_port_stub_add_connected_event_cb(stub_handle_,
      [](const char *sender, const char* instance, void *data) {
        RpcPortBase* p = static_cast<RpcPortBase*>(data);

        p->touch_stub_connected_event_cb_ = true;
      }, this);
  ASSERT_EQ(ret, 0);

  ret = rpc_port_stub_listen(stub_handle_);
  ASSERT_EQ(ret, 0);

  ret = rpc_port_proxy_add_connected_event_cb(proxy_handle_,
    [](const char *ep, const char *port_name, rpc_port_h port, void *data) {
      RpcPortBase* p = static_cast<RpcPortBase*>(data);

      p->touch_proxy_connected_event_cb_ = true;
    }, this);
  ASSERT_EQ(ret, 0);

  ret = rpc_port_proxy_connect(proxy_handle_, "TestApp", "test_port");
  ASSERT_EQ(ret, 0);

  ASSERT_TRUE(touch_proxy_connected_event_cb_);
  ASSERT_TRUE(touch_stub_connected_event_cb_);
}

TEST_F(RpcPortBase, rpc_port_proxy_event_reject) {
  int ret = rpc_port_stub_listen(stub_handle_);
  ASSERT_EQ(ret, 0);

  ret = rpc_port_proxy_add_rejected_event_cb(proxy_handle_,
    [](const char *ep, const char *port_name, void *data) {
      RpcPortBase* p = static_cast<RpcPortBase*>(data);

      p->touch_proxy_rejected_event_cb_ = true;
    }, this);
  ASSERT_EQ(ret, 0);

  ret = rpc_port_proxy_connect(proxy_handle_, "TestApp", "wrong_port");
  ASSERT_EQ(ret, 0);

  ASSERT_TRUE(touch_proxy_rejected_event_cb_);
}

TEST_F(RpcPortConnection, rpc_port_proxy_event_receive) {
  char res[] = "OK";
  char r_buf[256];

  ASSERT_NE(proxy_port_, nullptr);
  int ret = rpc_port_write(proxy_port_, res, sizeof(res));
  ASSERT_EQ(ret, 0);

  RunMainLoop();
  ASSERT_NE(stub_port_, nullptr);
  ASSERT_NE(stub_callback_port_, nullptr);

  ret = rpc_port_write(stub_callback_port_, res, sizeof(res));
  ASSERT_EQ(ret, 0);

  RunMainLoop();
  ASSERT_TRUE(touch_proxy_received_event_cb_);

  ret = rpc_port_read(proxy_callback_port_, r_buf, sizeof(res));
  ASSERT_EQ(ret, 0);
  ASSERT_STREQ(res, r_buf);

  touch_proxy_received_event_cb_ = false;
  ret = rpc_port_write(stub_callback_port_, res, sizeof(res));
  ASSERT_EQ(ret, 0);

  RunMainLoop();
  ASSERT_TRUE(touch_proxy_received_event_cb_);

  ret = rpc_port_read(proxy_callback_port_, r_buf, sizeof(res));
  ASSERT_EQ(ret, 0);
  ASSERT_STREQ(res, r_buf);
}

TEST_F(RpcPortConnection, rpc_port_read_write) {
  char buf[] = "test message";
  char res[] = "OK";
  char r_buf[256];

  ASSERT_NE(proxy_port_, nullptr);
  int ret = rpc_port_write(proxy_port_, buf, sizeof(buf));
  ASSERT_EQ(ret, 0);

  RunMainLoop();
  ASSERT_NE(stub_port_, nullptr);

  ret = rpc_port_read(stub_port_, r_buf, sizeof(buf));
  ASSERT_EQ(ret, 0);
  ASSERT_STREQ(buf, r_buf);

  ret = rpc_port_write(stub_port_, res, sizeof(res));
  ASSERT_EQ(ret, 0);

  ret = rpc_port_read(proxy_port_, r_buf, sizeof(res));
  ASSERT_EQ(ret, 0);
  ASSERT_STREQ("OK", r_buf);
}

TEST_F(RpcPortConnection, rpc_port_proxy_disconnected) {
  KillStub();
  RunMainLoop();

  ASSERT_TRUE(touch_proxy_disconnected_event_cb_);
}

TEST_F(RpcPortConnection, rpc_port_stub_disconnected) {
  KillProxy();
  RunMainLoop();

  ASSERT_TRUE(touch_stub_disconnected_event_cb_);
}
